from onegov.core.request import CoreRequest
from unittest.mock import patch


def test_views_payment_types(client):
    client.login_admin()

    edit = client.get('/payment-types')
    assert edit.form['normal'].value == '7.0'
    assert edit.form['spezial'].value == '8.5'

    edit.form['normal'] = 8
    edit.form['spezial'] = 9.9
    edited = edit.form.submit().follow()
    assert "geändert." in edited

    view = client.get('/payment-types')
    assert view.form['normal'].value == '8.0'
    assert view.form['spezial'].value == '9.9'


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_payment_types_permissions(mock_method, client):
    client.get('/payment-types', status=403)

    client.login_member()
    client.get('/payment-types', status=403)
    client.logout()

    client.login_editor()
    client.get('/payment-types', status=403)
    client.logout()

    client.login_admin()
    client.get('/payment-types')
    client.logout()



from datetime import date
from onegov.form import Form
from onegov.wtfs import _
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import DailyListBoxesAndForms
from wtforms import RadioField
from wtforms.fields.html5 import DateField
from wtforms.validators import InputRequired


class DailyListSelectionForm(Form):

    date = DateField(
        label=_("Date"),
        validators=[InputRequired()],
        default=date.today
    )

    type = RadioField(
        label=_("Daily list"),
        choices=[
            ('boxes', _("Boxes")),
            ('boxes_and_forms', _("Boxes and forms")),
        ],
        validators=[InputRequired()],
        default='boxes'
    )

    def get_model(self):
        if self.type.data == 'boxes':
            return DailyListBoxes(self.request.session, self.date.data)
        if self.type.data == 'boxes_and_forms':
            return DailyListBoxesAndForms(self.request.session, self.date.data)



from datetime import date
from datetime import timedelta
from dateutil.parser import parse
from onegov.core.orm.func import unaccent
from onegov.form import Form
from onegov.form.fields import ChosenSelectMultipleField
from onegov.form.fields import MultiCheckboxField
from onegov.form.fields import PreviewField
from onegov.wtfs import _
from onegov.wtfs.fields import HintField
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import PickupDate
from wtforms import HiddenField
from wtforms import IntegerField
from wtforms import RadioField
from wtforms import SelectField
from wtforms import StringField
from wtforms import TextAreaField
from wtforms_components import DateRange
from wtforms.fields.html5 import DateField
from wtforms.validators import InputRequired
from wtforms.validators import NumberRange
from wtforms.validators import Optional


def tomorrow():
    return date.today() + timedelta(days=1)


def coerce_date(value):
    if isinstance(value, str):
        return parse(value).date()
    return value


class AddScanJobForm(Form):

    callout = _("Fill in until 17.00 o'clock the evening before.")

    type = RadioField(
        label=_("Type"),
        choices=[('normal', _("Regular shipment"))],
        validators=[InputRequired()],
        default='normal'
    )

    type_hint = HintField(
        label="",
        macro='express_shipment_hint',
        depends_on=('type', 'express')
    )

    dispatch_date_normal = SelectField(
        label=_("Dispatch date"),
        choices=[],
        depends_on=('type', 'normal'),
        validators=[
            InputRequired(),
            DateRange(min=tomorrow, message=_("Date must be in the future."))
        ],
        coerce=coerce_date
    )

    dispatch_date_express = DateField(
        label=_("Dispatch date"),
        depends_on=('type', 'express'),
        validators=[
            InputRequired(),
            DateRange(min=tomorrow, message=_("Date must be in the future."))
        ],
        default=tomorrow
    )

    dispatch_boxes = IntegerField(
        label=_("Boxes"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3, 'clear': False},
    )
    dispatch_tax_forms_older = IntegerField(
        label=_("Tax forms (older)"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_tax_forms_last_year = IntegerField(
        label=_("Tax forms (previous year)"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_tax_forms_current_year = IntegerField(
        label=_("Tax forms"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_single_documents = IntegerField(
        label=_("Single documents"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3},
    )
    dispatch_note = TextAreaField(
        label=_("Note"),
        fieldset=_("Dispatch to the tax office"),
        render_kw={'rows': 5},
    )

    dispatch_cantonal_tax_office = IntegerField(
        label=_("Headquarters"),
        fieldset=_("Dispatch to the cantonal tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 6, 'clear': False},
    )
    dispatch_cantonal_scan_center = IntegerField(
        label=_("Scan center"),
        fieldset=_("Dispatch to the cantonal tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 6, 'clear': False},
    )

    @property
    def municipality_id(self):
        return self.request.identity.groupid or None

    @property
    def dispatch_date(self):
        if self.type.data == 'express':
            return self.dispatch_date_express.data
        return self.dispatch_date_normal.data

    def update_labels(self):
        year = date.today().year
        self.dispatch_tax_forms_older.label.text = _(
            "Tax forms until ${year}", mapping={'year': year - 2}
        )
        self.dispatch_tax_forms_last_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year - 1}
        )
        self.dispatch_tax_forms_current_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year}
        )

    def on_request(self):
        # Shipment types
        if self.request.has_role('editor'):
            self.type.choices = [
                ('normal', _("Regular shipment")),
                ('express', _("Express shipment"))
            ]

        # Dispatch dates
        query = self.request.session.query(PickupDate.date.label('date'))
        query = query.filter(
            PickupDate.municipality_id == self.municipality_id
        )
        query = query.filter(PickupDate.date > date.today())
        query = query.order_by(PickupDate.date)
        self.dispatch_date_normal.choices = [
            (r.date, f"{r.date:%d.%m.%Y}") for r in query
        ]

        # Labels
        self.update_labels()

    def update_model(self, model):
        model.municipality_id = self.request.identity.groupid
        model.type = self.type.data
        model.dispatch_date = self.dispatch_date
        for name in (
            'dispatch_boxes',
            'dispatch_tax_forms_current_year',
            'dispatch_tax_forms_last_year',
            'dispatch_tax_forms_older',
            'dispatch_single_documents',
            'dispatch_note',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
        ):
            setattr(model, name, getattr(self, name).data)


class EditScanJobForm(Form):

    callout = _("Fill in until 17.00 o'clock the evening before.")

    dispatch_boxes = IntegerField(
        label=_("Boxes"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3, 'clear': False},
    )
    dispatch_tax_forms_older = IntegerField(
        label=_("Tax forms (older)"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_tax_forms_last_year = IntegerField(
        label=_("Tax forms (previous year)"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_tax_forms_current_year = IntegerField(
        label=_("Tax forms"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_single_documents = IntegerField(
        label=_("Single documents"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3},
    )
    dispatch_note = TextAreaField(
        label=_("Note"),
        fieldset=_("Dispatch to the tax office"),
        render_kw={'rows': 5},
    )

    dispatch_cantonal_tax_office = IntegerField(
        label=_("Headquarters"),
        fieldset=_("Dispatch to the cantonal tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 6, 'clear': False},
    )
    dispatch_cantonal_scan_center = IntegerField(
        label=_("Scan center"),
        fieldset=_("Dispatch to the cantonal tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 6},
    )

    def update_labels(self):
        year = date.today().year
        self.dispatch_tax_forms_older.label.text = _(
            "Tax forms until ${year}", mapping={'year': year - 2}
        )
        self.dispatch_tax_forms_last_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year - 1}
        )
        self.dispatch_tax_forms_current_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year}
        )

    def on_request(self):
        self.update_labels()

    def update_model(self, model):
        for name in (
            'dispatch_boxes',
            'dispatch_tax_forms_current_year',
            'dispatch_tax_forms_last_year',
            'dispatch_tax_forms_older',
            'dispatch_single_documents',
            'dispatch_note',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
        ):
            setattr(model, name, getattr(self, name).data)

    def apply_model(self, model):
        for name in (
            'dispatch_boxes',
            'dispatch_tax_forms_current_year',
            'dispatch_tax_forms_last_year',
            'dispatch_tax_forms_older',
            'dispatch_single_documents',
            'dispatch_note',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
        ):
            getattr(self, name).data = getattr(model, name)

        self.callout = _(
            "Fill in until 17.00 o'clock the evening before ${date}.",
            mapping={'date': f"{model.dispatch_date:%d.%m.%Y}"}
        )


class UnrestrictedScanJobForm(Form):

    municipality_id = SelectField(
        label=_("Municipality"),
        choices=[],
        validators=[InputRequired()]
    )

    type = RadioField(
        label=_("Type"),
        choices=[
            ('normal', _("Regular shipment")),
            ('express', _("Express shipment"))
        ],
        validators=[InputRequired()],
        default='normal'
    )

    dispatch_date = DateField(
        label=_("Dispatch date"),
        validators=[InputRequired()],
        default=tomorrow
    )

    dispatch_date_hint = PreviewField(
        label=_("Regular dispatch dates"),
        fields=('municipality_id',),
        events=('change',),
        url=lambda meta: meta.request.link(
            meta.request.app.principal,
            name='dispatch-dates'
        )
    )

    dispatch_boxes = IntegerField(
        label=_("Boxes"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3, 'clear': False},
    )
    dispatch_tax_forms_older = IntegerField(
        label=_("Tax forms (older)"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_tax_forms_last_year = IntegerField(
        label=_("Tax forms (previous year)"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_tax_forms_current_year = IntegerField(
        label=_("Tax forms"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    dispatch_single_documents = IntegerField(
        label=_("Single documents"),
        fieldset=_("Dispatch to the tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3},
    )
    dispatch_note = TextAreaField(
        label=_("Note"),
        fieldset=_("Dispatch to the tax office"),
        render_kw={'rows': 5},
    )

    dispatch_cantonal_tax_office = IntegerField(
        label=_("Headquarters"),
        fieldset=_("Dispatch to the cantonal tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 6, 'clear': False},
    )
    dispatch_cantonal_scan_center = IntegerField(
        label=_("Scan center"),
        fieldset=_("Dispatch to the cantonal tax office"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 6},
    )

    return_date = DateField(
        label=_("Return date"),
        fieldset=_("Return to the municipality"),
        default=tomorrow,
        validators=[Optional()]
    )

    return_boxes = IntegerField(
        label=_("Boxes"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3, 'clear': False},
    )
    return_tax_forms_older = IntegerField(
        label=_("Tax forms (older)"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    return_tax_forms_last_year = IntegerField(
        label=_("Tax forms (previous year)"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    return_tax_forms_current_year = IntegerField(
        label=_("Tax forms"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    return_single_documents = IntegerField(
        label=_("Single documents"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3},

    )
    return_unscanned_tax_forms_older = IntegerField(
        label=_("Unscanned tax forms (older)"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'offset': 3, 'size': 2, 'clear': False},
    )
    return_unscanned_tax_forms_last_year = IntegerField(
        label=_("Unscanned tax forms (previous year)"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},
    )
    return_unscanned_tax_forms_current_year = IntegerField(
        label=_("Unscanned tax forms"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 2, 'clear': False},

    )
    return_unscanned_single_documents = IntegerField(
        label=_("Unscanned single documents"),
        fieldset=_("Return to the municipality"),
        validators=[Optional(), NumberRange(min=0)],
        render_kw={'size': 3},
    )
    return_note = TextAreaField(
        label=_("Note"),
        fieldset=_("Return to the municipality"),
        render_kw={'rows': 5},
    )

    def update_labels(self):
        year = date.today().year
        self.dispatch_tax_forms_older.label.text = _(
            "Tax forms until ${year}", mapping={'year': year - 2}
        )
        self.dispatch_tax_forms_last_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year - 1}
        )
        self.dispatch_tax_forms_current_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year}
        )
        self.return_tax_forms_older.label.text = _(
            "Tax forms until ${year}", mapping={'year': year - 2}
        )
        self.return_tax_forms_last_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year - 1}
        )
        self.return_tax_forms_current_year.label.text = _(
            "Tax forms ${year}", mapping={'year': year}
        )
        self.return_unscanned_tax_forms_older.label.text = _(
            "Unscanned tax forms until ${year}", mapping={'year': year - 2}
        )
        self.return_unscanned_tax_forms_last_year.label.text = _(
            "Unscanned tax forms ${year}", mapping={'year': year - 1}
        )
        self.return_unscanned_tax_forms_current_year.label.text = _(
            "Unscanned tax forms ${year}", mapping={'year': year}
        )

    def on_request(self):
        query = self.request.session.query(
            Municipality.id.label('id'),
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number'),
        )
        query = query.order_by(unaccent(Municipality.name))
        self.municipality_id.choices = [
            (r.id.hex, f"{r.name} ({r.bfs_number})") for r in query
        ]

        self.update_labels()

    def update_model(self, model):
        for name in (
            'municipality_id',
            'type',
            'dispatch_date',
            'dispatch_boxes',
            'dispatch_tax_forms_current_year',
            'dispatch_tax_forms_last_year',
            'dispatch_tax_forms_older',
            'dispatch_single_documents',
            'dispatch_note',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
            'return_date',
            'return_boxes',
            'return_tax_forms_current_year',
            'return_tax_forms_last_year',
            'return_tax_forms_older',
            'return_single_documents',
            'return_unscanned_tax_forms_current_year',
            'return_unscanned_tax_forms_last_year',
            'return_unscanned_tax_forms_older',
            'return_unscanned_single_documents',
            'return_note',
        ):
            setattr(model, name, getattr(self, name).data)

    def apply_model(self, model):
        self.municipality_id.data = model.municipality_id.hex
        for name in (
            'type',
            'dispatch_date',
            'dispatch_boxes',
            'dispatch_tax_forms_current_year',
            'dispatch_tax_forms_last_year',
            'dispatch_tax_forms_older',
            'dispatch_single_documents',
            'dispatch_note',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
            'return_date',
            'return_boxes',
            'return_tax_forms_current_year',
            'return_tax_forms_last_year',
            'return_tax_forms_older',
            'return_single_documents',
            'return_unscanned_tax_forms_current_year',
            'return_unscanned_tax_forms_last_year',
            'return_unscanned_tax_forms_older',
            'return_unscanned_single_documents',
            'return_note',
        ):
            getattr(self, name).data = getattr(model, name)


class ScanJobsForm(Form):

    sort_by = HiddenField()
    sort_order = HiddenField()

    from_date = DateField(
        label=_("Start date"),
        fieldset=_("Filter")
    )

    to_date = DateField(
        label=_("End date"),
        fieldset=_("Filter")
    )

    type = MultiCheckboxField(
        label=_("Type"),
        fieldset=_("Filter"),
        choices=[
            ('normal', _("Regular shipment")),
            ('express', _("Express shipment"))
        ]
    )

    term = StringField(
        label=_("Term"),
        fieldset=_("Filter")
    )

    def on_request(self):
        if hasattr(self, 'csrf_token'):
            self.delete_field('csrf_token')

    def select_all(self, name):
        field = getattr(self, name)
        if not field.data:
            field.data = list(next(zip(*field.choices)))

    def apply_model(self, model):
        self.from_date.data = model.from_date
        self.to_date.data = model.to_date
        self.type.data = model.type
        self.term.data = model.term
        self.sort_by.data = model.sort_by
        self.sort_order.data = model.sort_order

        # default unselected checkboxes to all choices
        self.select_all('type')


class UnrestrictedScanJobsForm(ScanJobsForm):

    municipality_id = ChosenSelectMultipleField(
        label=_("Municipality"),
        fieldset=_("Filter")
    )

    def on_request(self):
        super().on_request()
        query = self.request.session.query(
            Municipality.id.label('id'),
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number')
        )
        query = query.order_by(unaccent(Municipality.name))
        self.municipality_id.choices = [
            (r.id.hex, f"{r.name} ({r.bfs_number})") for r in query
        ]

    def apply_model(self, model):
        super().apply_model(model)
        self.municipality_id.data = model.municipality_id



import click

from onegov.core.cli import abort
from onegov.core.cli import command_group
from onegov.core.cli import pass_group_context
from sqlalchemy import create_engine


cli = command_group()


@cli.command(context_settings={'creates_path': True})
@pass_group_context
def add(group_context):
    """ Adds an instance to the database. For example:

        onegov-wtfs --select '/onegov_wtfs/wtfs' add

    """

    def add_instance(request, app):
        app.cache.invalidate()
        app.add_initial_content()
        click.echo("Instance was created successfully")

    return add_instance


@cli.command()
@pass_group_context
def delete(group_context):
    """ Deletes an instance from the database. For example:

        onegov-wtfs --select '/onegov_wtfs/wtfs' delete

    """

    def delete_instance(request, app):

        confirmation = "Do you really want to DELETE {}?".format(app.schema)

        if not click.confirm(confirmation):
            abort("Deletion process aborted")

        assert app.has_database_connection
        assert app.session_manager.is_valid_schema(app.schema)

        dsn = app.session_manager.dsn
        app.session_manager.session().close_all()
        app.session_manager.dispose()

        engine = create_engine(dsn)
        engine.execute('DROP SCHEMA "{}" CASCADE'.format(app.schema))
        engine.raw_connection().invalidate()
        engine.dispose()

        click.echo("Instance was deleted successfully")

    return delete_instance



from onegov.core.request import CoreRequest
from onegov.core.utils import module_path
from pytest import mark
from unittest.mock import patch
from webtest.forms import Upload


@mark.parametrize("pdf_1, pdf_2", [(
    module_path('onegov.wtfs', 'tests/fixtures/example_1.pdf'),
    module_path('onegov.wtfs', 'tests/fixtures/example_2.pdf')
)])
def test_views_user_manual(client, pdf_1, pdf_2):
    with open(pdf_1, 'rb') as file:
        pdf_1 = file.read()
    with open(pdf_2, 'rb') as file:
        pdf_2 = file.read()

    client.login_admin()

    view = client.get('/user-manual')
    client.get('http://localhost/user-manual/pdf', status=503)
    assert "Noch kein Benutzerhandbuch vorhanden." in view

    add = view.click("Bearbeiten")
    add.form['pdf'] = Upload(f'Handbuch.pdf', pdf_1, 'application/pdf')
    added = add.form.submit().follow()
    assert "Benutzerhandbuch geändert." in added
    assert "Benutzerhandbuch (PDF, 8.1 kB)." in added
    assert client.get('http://localhost/user-manual/pdf').body == pdf_1

    # Edit
    edited = view.click("Bearbeiten")
    assert "user_manual.pdf (8.1 kB)" in str(edited.form.html)
    edited.form.get('pdf', 0).select('replace')
    edited.form.get('pdf', 1).value = Upload(f'xx', pdf_2, 'application/pdf')
    editeded = edited.form.submit().follow()
    assert "Benutzerhandbuch geändert." in editeded
    assert "Benutzerhandbuch (PDF, 9.1 kB)." in editeded
    assert client.get('http://localhost/user-manual/pdf').body == pdf_2

    # Delete
    edited = view.click("Bearbeiten")
    assert "user_manual.pdf (9.1 kB)" in str(edited.form.html)
    edited.form.get('pdf', 0).select('delete')
    editeded = edited.form.submit().follow()
    assert "Benutzerhandbuch geändert." in editeded
    assert "Benutzerhandbuch (PDF, 9.1 kB)." not in editeded
    assert client.get('http://localhost/user-manual/pdf', status=503)


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_user_manual_permissions(mock_method, client):
    client.get('/user-manual', status=403)
    client.get('/user-manual/edit', status=403)

    client.login_member()
    client.get('/user-manual')
    client.get('/user-manual/edit', status=403)
    client.logout()

    client.login_editor()
    client.get('/user-manual')
    client.get('/user-manual/edit', status=403)
    client.logout()

    client.login_admin()
    client.get('/user-manual')
    client.get('/user-manual/edit')
    client.logout()



from onegov.core.orm import Base
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.wtfs import _
from onegov.wtfs.models.municipality import Municipality
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from uuid import uuid4


class ScanJob(Base, TimestampMixin, ContentMixin):
    """ A tax form scan job date. """

    __tablename__ = 'wtfs_scan_jobs'

    #: the id of the db record (only relevant internally)
    id = Column(UUID, primary_key=True, default=uuid4)

    #: The municipality that owns the scan job.
    municipality_id = Column(UUID, ForeignKey(Municipality.id), nullable=False)
    municipality = relationship(
        Municipality,
        backref=backref(
            'scan_jobs',
            lazy='dynamic',
            order_by='ScanJob.dispatch_date'
        )
    )

    #: The scan job type.
    type = Column(
        Enum('normal', 'express', name='wtfs_scan_job_type'),
        nullable=False
    )

    #: The delivery number.
    delivery_number = Column(
        Integer,
        nullable=False
    )

    dispatch_date = Column(Date, nullable=False)
    dispatch_note = Column(Text, nullable=True)
    dispatch_boxes = Column(Integer, nullable=True)

    dispatch_tax_forms_current_year = Column(Integer, nullable=True)
    dispatch_tax_forms_last_year = Column(Integer, nullable=True)
    dispatch_tax_forms_older = Column(Integer, nullable=True)

    @hybrid_property
    def dispatch_tax_forms(self):
        return (
            (self.dispatch_tax_forms_current_year or 0)
            + (self.dispatch_tax_forms_last_year or 0)
            + (self.dispatch_tax_forms_older or 0)
        ) or None

    @dispatch_tax_forms.expression
    def dispatch_tax_forms(cls):
        return (
            func.coalesce(cls.dispatch_tax_forms_current_year, 0)
            + func.coalesce(cls.dispatch_tax_forms_last_year, 0)
            + func.coalesce(cls.dispatch_tax_forms_older, 0)
        )

    dispatch_single_documents = Column(Integer, nullable=True)
    dispatch_cantonal_tax_office = Column(Integer, nullable=True)
    dispatch_cantonal_scan_center = Column(Integer, nullable=True)

    return_date = Column(Date, nullable=True)
    return_note = Column(Text, nullable=True)
    return_boxes = Column(Integer, nullable=True)

    return_tax_forms_current_year = Column(Integer, nullable=True)
    return_tax_forms_last_year = Column(Integer, nullable=True)
    return_tax_forms_older = Column(Integer, nullable=True)

    @hybrid_property
    def return_tax_forms(self):
        return (
            (self.return_tax_forms_current_year or 0)
            + (self.return_tax_forms_last_year or 0)
            + (self.return_tax_forms_older or 0)
        ) or None

    @return_tax_forms.expression
    def return_tax_forms(cls):
        return (
            func.coalesce(cls.return_tax_forms_current_year, 0)
            + func.coalesce(cls.return_tax_forms_last_year, 0)
            + func.coalesce(cls.return_tax_forms_older, 0)
        )

    return_single_documents = Column(Integer, nullable=True)

    return_unscanned_tax_forms_current_year = Column(Integer, nullable=True)
    return_unscanned_tax_forms_last_year = Column(Integer, nullable=True)
    return_unscanned_tax_forms_older = Column(Integer, nullable=True)

    @hybrid_property
    def return_unscanned_tax_forms(self):
        return (
            (self.return_unscanned_tax_forms_current_year or 0)
            + (self.return_unscanned_tax_forms_last_year or 0)
            + (self.return_unscanned_tax_forms_older or 0)
        ) or None

    @return_unscanned_tax_forms.expression
    def return_unscanned_tax_forms(cls):
        return (
            func.coalesce(cls.return_unscanned_tax_forms_current_year, 0)
            + func.coalesce(cls.return_unscanned_tax_forms_last_year, 0)
            + func.coalesce(cls.return_unscanned_tax_forms_older, 0)
        )

    return_unscanned_single_documents = Column(Integer, nullable=True)

    @hybrid_property
    def return_scanned_tax_forms_current_year(self):
        return (
            (self.return_tax_forms_current_year or 0)
            - (self.return_unscanned_tax_forms_current_year or 0)
        ) or None

    @return_scanned_tax_forms_current_year.expression
    def return_scanned_tax_forms_current_year(cls):
        return (
            func.coalesce(cls.return_tax_forms_current_year, 0)
            - func.coalesce(cls.return_unscanned_tax_forms_current_year, 0)
        )

    @hybrid_property
    def return_scanned_tax_forms_last_year(self):
        return (
            (self.return_tax_forms_last_year or 0)
            - (self.return_unscanned_tax_forms_last_year or 0)
        ) or None

    @return_scanned_tax_forms_last_year.expression
    def return_scanned_tax_forms_last_year(cls):
        return (
            func.coalesce(cls.return_tax_forms_last_year, 0)
            - func.coalesce(cls.return_unscanned_tax_forms_last_year, 0)
        )

    @hybrid_property
    def return_scanned_tax_forms_older(self):
        return (
            (self.return_tax_forms_older or 0)
            - (self.return_unscanned_tax_forms_older or 0)
        ) or None

    @return_scanned_tax_forms_older.expression
    def return_scanned_tax_forms_older(cls):
        return (
            func.coalesce(cls.return_tax_forms_older, 0)
            - func.coalesce(cls.return_unscanned_tax_forms_older, 0)
        )

    @hybrid_property
    def return_scanned_tax_forms(self):
        return (
            (self.return_tax_forms or 0)
            - (self.return_unscanned_tax_forms or 0)
        ) or None

    @return_scanned_tax_forms.expression
    def return_scanned_tax_forms(cls):
        return (
            func.coalesce(cls.return_tax_forms, 0)
            - func.coalesce(cls.return_unscanned_tax_forms, 0)
        )

    @hybrid_property
    def return_scanned_single_documents(self):
        return (
            (self.return_single_documents or 0)
            - (self.return_unscanned_single_documents or 0)
        ) or None

    @return_scanned_single_documents.expression
    def return_scanned_single_documents(cls):
        return (
            func.coalesce(cls.return_single_documents, 0)
            - func.coalesce(cls.return_unscanned_single_documents, 0)
        )

    @property
    def title(self):
        return _(
            "Scan job no. ${number}",
            mapping={'number': self.delivery_number}
        )



from onegov.form import Form
from wtforms import FloatField
from wtforms.validators import InputRequired


class PaymentTypesForm(Form):

    @staticmethod
    def get_form_class(model, request):
        form = PaymentTypesForm
        for payment_type in model.query().all():
            field = FloatField(
                payment_type.name.capitalize(),
                validators=[InputRequired()]
            )
            setattr(form, payment_type.name, field)
        return form

    def update_model(self, model):
        for field in self:
            payment_type = model.query().filter_by(name=field.name).first()
            if payment_type:
                payment_type.price_per_quantity = field.data

    def apply_model(self, model):
        for field in self:
            payment_type = model.query().filter_by(name=field.name).first()
            if payment_type:
                field.data = payment_type.price_per_quantity



from morepath import redirect
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.forms import DeleteMunicipalityDatesForm
from onegov.wtfs.forms import ImportMunicipalityDataForm
from onegov.wtfs.forms import MunicipalityForm
from onegov.wtfs.layouts import AddMunicipalityLayout
from onegov.wtfs.layouts import DeleteMunicipalityDatesLayout
from onegov.wtfs.layouts import EditMunicipalityLayout
from onegov.wtfs.layouts import ImportMunicipalityDataLayout
from onegov.wtfs.layouts import MunicipalitiesLayout
from onegov.wtfs.layouts import MunicipalityLayout
from onegov.wtfs.models import Municipality
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import ViewModel


@WtfsApp.html(
    model=MunicipalityCollection,
    template='municipalities.pt',
    permission=ViewModel
)
def view_municipalities(self, request):
    """ View the list of municipalities. """

    layout = MunicipalitiesLayout(self, request)

    return {
        'layout': layout,
    }


@WtfsApp.form(
    model=MunicipalityCollection,
    name='import-data',
    template='form.pt',
    permission=EditModel,
    form=ImportMunicipalityDataForm
)
def import_municipality_data(self, request, form):
    """ Import municipality data. """

    layout = ImportMunicipalityDataLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Municipality data imported."), 'success')
        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Import"),
        'cancel': layout.cancel_url
    }


@WtfsApp.form(
    model=MunicipalityCollection,
    name='add',
    template='form.pt',
    permission=AddModel,
    form=MunicipalityForm
)
def add_municipality(self, request, form):
    """ Create a new municipality. """

    layout = AddMunicipalityLayout(self, request)

    if form.submitted(request):
        municipality = Municipality()
        form.update_model(municipality)
        request.session.add(municipality)
        request.message(_("Municipality added."), 'success')
        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }


@WtfsApp.html(
    model=Municipality,
    template='municipality.pt',
    permission=ViewModel
)
def view_municipality(self, request):
    """ View a single municipality. """

    layout = MunicipalityLayout(self, request)

    return {
        'layout': layout,
    }


@WtfsApp.form(
    model=Municipality,
    name='edit',
    template='form.pt',
    permission=EditModel,
    form=MunicipalityForm
)
def edit_municipality(self, request, form):
    """ Edit a municipality. """

    layout = EditMunicipalityLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Municipality modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.form(
    model=Municipality,
    name='delete-dates',
    template='form.pt',
    permission=EditModel,
    form=DeleteMunicipalityDatesForm
)
def delete_municipality_dates(self, request, form):
    """ Delete a range of pick-up dates of a municipality. """

    layout = DeleteMunicipalityDatesLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Pick-up dates deleted."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Delete"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.view(
    model=Municipality,
    request_method='DELETE',
    permission=DeleteModel
)
def delete_municipality(self, request):
    """ Delete a municipality. """

    request.assert_valid_csrf_token()
    MunicipalityCollection(request.session).delete(self)
    request.message(_("Municipality deleted."), 'success')



from datetime import date
from onegov.core.orm.func import unaccent
from onegov.form import Form
from onegov.wtfs import _
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import ReportBoxes
from onegov.wtfs.models import ReportBoxesAndForms
from onegov.wtfs.models import ReportBoxesAndFormsByDelivery
from onegov.wtfs.models import ReportFormsByMunicipality
from wtforms import RadioField
from wtforms import SelectField
from wtforms.fields.html5 import DateField
from wtforms.validators import InputRequired


class ReportSelectionForm(Form):

    start = DateField(
        label=_("Start date"),
        validators=[InputRequired()],
        default=date.today
    )

    end = DateField(
        label=_("End date"),
        validators=[InputRequired()],
        default=date.today
    )

    report_type = RadioField(
        label=_("Report"),
        choices=[
            ('boxes', _("Report boxes")),
            ('boxes_and_forms', _("Report boxes and forms")),
            ('forms', _("Report forms")),
            ('delivery', _("Report boxes and forms by delivery")),
        ],
        validators=[InputRequired()],
        default='boxes'
    )

    scan_job_type = RadioField(
        label=_("Type"),
        choices=[
            ('normal', _("Regular shipment")),
            ('express', _("Express shipment")),
            ('all', _("Regular and express shipment"))
        ],
        depends_on=('report_type', '!boxes'),
        validators=[InputRequired()],
        default='all'
    )

    municipality_id = SelectField(
        label=_("Municipality"),
        choices=[],
        validators=[InputRequired()],
        depends_on=(
            'report_type', '!boxes',
            'report_type', '!boxes_and_forms'
        ),
    )

    def on_request(self):
        query = self.request.session.query(
            Municipality.id.label('id'),
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number')
        )
        query = query.order_by(unaccent(Municipality.name))
        self.municipality_id.choices = [
            (r.id.hex, f"{r.name} ({r.bfs_number})") for r in query
        ]

    def get_model(self):
        if self.report_type.data == 'boxes':
            return ReportBoxes(
                self.request.session,
                start=self.start.data,
                end=self.end.data
            )
        if self.report_type.data == 'boxes_and_forms':
            return ReportBoxesAndForms(
                self.request.session,
                start=self.start.data,
                end=self.end.data,
                type=self.scan_job_type.data
            )
        if self.report_type.data == 'forms':
            return ReportFormsByMunicipality(
                self.request.session,
                start=self.start.data,
                end=self.end.data,
                type=self.scan_job_type.data,
                municipality_id=self.municipality_id.data
            )
        if self.report_type.data == 'delivery':
            return ReportBoxesAndFormsByDelivery(
                self.request.session,
                start=self.start.data,
                end=self.end.data,
                type=self.scan_job_type.data,
                municipality_id=self.municipality_id.data
            )



from dateutil.parser import parse
from onegov.form.fields import UploadField
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.wtfs import _


class MunicipalityDataUploadField(UploadField):
    """ An upload field containg municipality data. """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('validators', [])
        kwargs['validators'].append(WhitelistedMimeType({'text/plain', }))
        kwargs['validators'].append(FileSizeLimit(10 * 1024 * 1024))

        kwargs.setdefault('render_kw', {})
        kwargs['render_kw']['force_simple'] = True

        super().__init__(*args, **kwargs)

    def post_validate(self, form, validation_stopped):
        errors = []
        data = {}

        if not self.raw_data:
            raise ValueError(_("No data"))

        lines = self.raw_data[0].file.read().decode('cp1252').split('\r\n')
        for line_number, line in enumerate(lines):
            if not line.strip():
                continue
            try:
                parts = [part.strip() for part in line.split(';')]
                parts = [part for part in parts if part]
                bfs_number = int(parts[1])
                dates = [parse(d, dayfirst=True).date() for d in parts[4:]]
            except (IndexError, TypeError, ValueError):
                errors.append(line_number)
            else:
                data[bfs_number] = {'dates': dates}

        if errors:
            raise ValueError(_(
                "Some rows contain invalid values: ${errors}.",
                mapping={'errors': ', '.join((str(e) for e in errors))}
            ))

        self.data = data



from onegov.wtfs.layouts.daily_list import DailyListBoxesAndFormsLayout
from onegov.wtfs.layouts.daily_list import DailyListBoxesLayout
from onegov.wtfs.layouts.daily_list import DailyListLayout
from onegov.wtfs.layouts.default import DefaultLayout
from onegov.wtfs.layouts.invoice import InvoiceLayout
from onegov.wtfs.layouts.mail import MailLayout
from onegov.wtfs.layouts.municipality import AddMunicipalityLayout
from onegov.wtfs.layouts.municipality import DeleteMunicipalityDatesLayout
from onegov.wtfs.layouts.municipality import EditMunicipalityLayout
from onegov.wtfs.layouts.municipality import ImportMunicipalityDataLayout
from onegov.wtfs.layouts.municipality import MunicipalitiesLayout
from onegov.wtfs.layouts.municipality import MunicipalityLayout
from onegov.wtfs.layouts.notification import AddNotificationLayout
from onegov.wtfs.layouts.notification import EditNotificationLayout
from onegov.wtfs.layouts.notification import NotificationLayout
from onegov.wtfs.layouts.notification import NotificationsLayout
from onegov.wtfs.layouts.payment_type import PaymentTypesLayout
from onegov.wtfs.layouts.report import ReportBoxesAndFormsByDeliveryLayout
from onegov.wtfs.layouts.report import ReportBoxesAndFormsLayout
from onegov.wtfs.layouts.report import ReportBoxesLayout
from onegov.wtfs.layouts.report import ReportFormsByMunicipalityLayout
from onegov.wtfs.layouts.report import ReportLayout
from onegov.wtfs.layouts.scan_job import AddScanJobLayout
from onegov.wtfs.layouts.scan_job import DeliveryNoteLayout
from onegov.wtfs.layouts.scan_job import EditScanJobLayout
from onegov.wtfs.layouts.scan_job import ScanJobLayout
from onegov.wtfs.layouts.scan_job import ScanJobsLayout
from onegov.wtfs.layouts.user import AddUserLayout
from onegov.wtfs.layouts.user import EditUserLayout
from onegov.wtfs.layouts.user import UserLayout
from onegov.wtfs.layouts.user import UsersLayout
from onegov.wtfs.layouts.user_manual import EditUserManualLayout
from onegov.wtfs.layouts.user_manual import UserManualLayout


__all__ = (
    'AddMunicipalityLayout',
    'AddNotificationLayout',
    'AddScanJobLayout',
    'AddUserLayout',
    'DailyListBoxesAndFormsLayout',
    'DailyListBoxesLayout',
    'DailyListLayout',
    'DefaultLayout',
    'DeleteMunicipalityDatesLayout',
    'DeliveryNoteLayout',
    'EditMunicipalityLayout',
    'EditNotificationLayout',
    'EditScanJobLayout',
    'EditUserLayout',
    'EditUserManualLayout',
    'ImportMunicipalityDataLayout',
    'InvoiceLayout',
    'MailLayout',
    'MunicipalitiesLayout',
    'MunicipalityLayout',
    'NotificationLayout',
    'NotificationsLayout',
    'PaymentTypesLayout',
    'ReportBoxesAndFormsByDeliveryLayout',
    'ReportBoxesAndFormsLayout',
    'ReportBoxesLayout',
    'ReportFormsByMunicipalityLayout',
    'ReportLayout',
    'ScanJobLayout',
    'ScanJobsLayout',
    'UserLayout',
    'UserManualLayout',
    'UsersLayout',
)



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.wtfs.models.municipality import Municipality
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import ForeignKey
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from uuid import uuid4


class PickupDate(Base, TimestampMixin):
    """ A pick-up date. """

    __tablename__ = 'wtfs_pickup_dates'

    #: the id of the db record (only relevant internally)
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the date
    date = Column(Date, nullable=False)

    #: the municipality.
    municipality_id = Column(UUID, ForeignKey(Municipality.id), nullable=False)
    municipality = relationship(
        Municipality,
        backref=backref(
            'pickup_dates',
            lazy='dynamic',
            order_by='PickupDate.date'
        )
    )



from datetime import date
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.collections import NotificationCollection
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.collections import ScanJobCollection
from uuid import uuid4


def test_payment_types(session):
    payment_types = PaymentTypeCollection(session)
    payment_type = payment_types.add(
        name='normal',
        _price_per_quantity=700
    )

    assert payment_type.name == 'normal'
    assert payment_type._price_per_quantity == 700
    assert payment_type.price_per_quantity == 7.0


def test_municipalities(session):
    municipalities = MunicipalityCollection(session)
    municipalities.add(
        name='Boppelsen',
        bfs_number=82,
    )
    municipalities.add(
        name='Adlikon',
        bfs_number=21,
    )

    assert [(m.name, m.bfs_number) for m in municipalities.query()] == [
        ('Adlikon', 21),
        ('Boppelsen', 82)
    ]

    # Import data
    data = {
        21: {'dates': [date(2019, 1, 1), date(2019, 1, 7)]},
        241: {'dates': [date(2019, 1, 2), date(2019, 1, 8)]},
        82: {'dates': [date(2019, 1, 4), date(2019, 1, 10)]}
    }
    municipalities.import_data(data)
    assert [
        (m.bfs_number, [d.date for d in m.pickup_dates])
        for m in municipalities.query()
    ] == [
        (21, [date(2019, 1, 1), date(2019, 1, 7)]),
        (82, [date(2019, 1, 4), date(2019, 1, 10)])
    ]

    municipalities.add(
        name='Aesch',
        bfs_number=241,
    )
    data = {
        21: {'dates': [date(2019, 1, 7), date(2019, 1, 8)]},
        241: {'dates': [date(2019, 1, 2), date(2019, 1, 8)]},
        82: {'dates': []}
    }
    municipalities.import_data(data)
    assert [
        (m.bfs_number, [d.date for d in m.pickup_dates])
        for m in municipalities.query()
    ] == [
        (21, [date(2019, 1, 1), date(2019, 1, 7), date(2019, 1, 8)]),
        (241, [date(2019, 1, 2), date(2019, 1, 8)]),
        (82, [date(2019, 1, 4), date(2019, 1, 10)])
    ]


def test_notifications(session):
    notifications = NotificationCollection(session)
    notification = notifications.add(
        title="Lorem ipsum",
        text="Lorem ipsum dolor sit amet.",
        channel_id="wtfs",
        owner="admin"
    )

    assert notification.title == "Lorem ipsum"
    assert notification.text == "Lorem ipsum dolor sit amet."
    assert notification.channel_id == "wtfs"
    assert notification.owner == "admin"
    assert notification.type == "wtfs_notification"


def test_scan_jobs(session):
    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name='Boppelsen', bfs_number=82)

    scan_jobs = ScanJobCollection(session)
    for day in (2, 1, 4, 3):
        scan_jobs.add(
            type='normal',
            municipality_id=municipality.id,
            dispatch_date=date(2019, 1, day)
        )

    assert [s.dispatch_date.day for s in scan_jobs.query()] == [4, 3, 2, 1]
    assert [s.delivery_number for s in scan_jobs.query()] == [3, 4, 1, 2]

    municipality = municipalities.add(name='Adlikon', bfs_number=21)
    scan_jobs.add(
        type='normal',
        municipality_id=municipality.id,
        dispatch_date=date(2019, 1, 5)
    )
    assert [s.delivery_number for s in scan_jobs.query()] == [1, 3, 4, 1, 2]

    assert scan_jobs.next_delivery_number(None) == 1
    assert scan_jobs.next_delivery_number(municipality.id) == 2


def test_scan_jobs_default():
    scan_jobs = ScanJobCollection(
        session=1,
        page=3,
        from_date=4,
        to_date=5,
        type=6,
        municipality_id=7,
        sort_by=8,
        sort_order=9
    )
    assert scan_jobs.session == 1
    assert scan_jobs.page == 3
    assert scan_jobs.from_date == 4
    assert scan_jobs.to_date == 5
    assert scan_jobs.type == 6
    assert scan_jobs.municipality_id == 7
    assert scan_jobs.sort_by == 8
    assert scan_jobs.sort_order == 9

    scan_jobs = scan_jobs.default()
    assert scan_jobs.session == 1
    assert scan_jobs.page is None
    assert scan_jobs.from_date is None
    assert scan_jobs.to_date is None
    assert scan_jobs.type is None
    assert scan_jobs.municipality_id is None
    assert scan_jobs.sort_by is None
    assert scan_jobs.sort_order is None


def test_scan_jobs_pagination(session):
    municipality_id = uuid4()
    MunicipalityCollection(session).add(
        id=municipality_id,
        name='Adlikon',
        bfs_number=21
    )

    scan_jobs = ScanJobCollection(session)

    assert scan_jobs.pages_count == 0
    assert scan_jobs.batch == []
    assert scan_jobs.page_index == 0
    assert scan_jobs.offset == 0
    assert scan_jobs.previous is None
    assert scan_jobs.next is None
    assert scan_jobs.page_by_index(0) == scan_jobs

    for day in range(1, 27):
        scan_jobs.add(
            type='express',
            dispatch_date=date(2019, 1, day),
            municipality_id=municipality_id,
        )

    scan_jobs = ScanJobCollection(session)
    assert scan_jobs.query().count() == 26
    assert scan_jobs.subset_count == 26
    assert scan_jobs.pages_count == 2
    assert len(scan_jobs.batch) == 20
    assert scan_jobs.page_index == 0
    assert scan_jobs.offset == 0
    assert scan_jobs.previous is None
    assert scan_jobs.next == scan_jobs.page_by_index(1)
    assert scan_jobs.page_by_index(0) == scan_jobs

    assert scan_jobs.next.page_index == 1
    assert len(scan_jobs.next.batch) == 6
    assert scan_jobs.next.previous == scan_jobs


def test_scan_jobs_query(session):
    municipalities = MunicipalityCollection(session)
    municipalities = {
        'Adlikon': municipalities.add(name='Adlikon', bfs_number=21),
        'Aesch': municipalities.add(name='Aesch', bfs_number=241)
    }

    def add(name, dispatch_date, type, dispatch_note, return_note):
        ScanJobCollection(session).add(
            dispatch_date=dispatch_date,
            type=type,
            municipality_id=municipalities[name].id,
            dispatch_note=dispatch_note,
            return_note=return_note,
        )

    add('Adlikon', date(2019, 1, 6), 'normal', None, 'bc')
    add('Adlikon', date(2019, 1, 5), 'express', 'abc', None)
    add('Adlikon', date(2019, 1, 4), 'express', '', '')
    add('Aesch', date(2019, 1, 4), 'normal', None, 'Lorem ipsum')
    add('Aesch', date(2019, 1, 3), 'express', 'Lorem Ipsum', None)
    add('Aesch', date(2019, 1, 2), 'express', None, 'loremipsum')

    def count(**kwargs):
        return ScanJobCollection(session, **kwargs).query().count()

    assert count() == 6

    assert count(from_date=date(2019, 1, 1)) == 6
    assert count(from_date=date(2019, 1, 2)) == 6
    assert count(from_date=date(2019, 1, 3)) == 5
    assert count(from_date=date(2019, 1, 5)) == 2
    assert count(from_date=date(2019, 1, 6)) == 1
    assert count(from_date=date(2019, 1, 7)) == 0

    assert count(to_date=date(2019, 1, 1)) == 0
    assert count(to_date=date(2019, 1, 2)) == 1
    assert count(to_date=date(2019, 1, 3)) == 2
    assert count(to_date=date(2019, 1, 5)) == 5
    assert count(to_date=date(2019, 1, 6)) == 6
    assert count(to_date=date(2019, 1, 7)) == 6

    assert count(from_date=date(2019, 1, 3), to_date=date(2019, 1, 4)) == 3
    assert count(from_date=date(2019, 1, 4), to_date=date(2019, 1, 3)) == 0

    assert count(type=[]) == 6
    assert count(type=['normal']) == 2
    assert count(type=['express']) == 4
    assert count(type=['normal', 'express']) == 6

    assert count(municipality_id=[]) == 6
    assert count(municipality_id=[municipalities['Adlikon'].id]) == 3
    assert count(municipality_id=[municipalities['Adlikon'].id,
                                  municipalities['Aesch'].id]) == 6

    assert count(term='abc') == 1
    assert count(term='bc') == 2
    assert count(term='C') == 2
    assert count(term='LOREM IPSUM') == 2
    assert count(term='ips') == 3

    assert count(
        municipality_id=[municipalities['Adlikon'].id],
        type=['express'],
        from_date=date(2019, 1, 4),
        term='a'
    ) == 1


def test_scan_jobs_order(session):
    scan_jobs = ScanJobCollection(session)
    for day, name in ((3, 'First'), (1, 'Śecond'), (2, 'Phirrrrd')):
        municipality = MunicipalityCollection(session).add(
            name=name,
            bfs_number=day,
        )
        scan_jobs.add(
            dispatch_date=date(2019, 1, day),
            type='normal',
            municipality_id=municipality.id,
        )
    scan_jobs.add(
        dispatch_date=date(2019, 1, 4),
        type='normal',
        municipality_id=municipality.id,
    )

    def collection(sort_by=None, sort_order=None):
        return ScanJobCollection(
            session, sort_by=sort_by, sort_order=sort_order
        )

    scan_jobs = collection()
    assert scan_jobs.sort_order_by_key('dispatch_date') == 'descending'
    assert scan_jobs.sort_order_by_key('delivery_number') == 'unsorted'
    assert scan_jobs.sort_order_by_key('municipality_id') == 'unsorted'
    assert scan_jobs.sort_order_by_key('invalid') == 'unsorted'

    scan_jobs = collection('', '')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'

    scan_jobs = collection('xx', 'yy')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'

    scan_jobs = collection('dispatch_date', 'yy')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'

    scan_jobs = collection('xx', 'ascending')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'

    scan_jobs = collection('delivery_number', 'yy')
    assert scan_jobs.current_sort_by == 'delivery_number'
    assert scan_jobs.current_sort_order == 'ascending'

    scan_jobs = collection()
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'
    assert 'dispatch_date' in str(scan_jobs.order_by)
    assert 'DESC' in str(scan_jobs.order_by)
    assert [s.dispatch_date.day for s in scan_jobs.query()] == [4, 3, 2, 1]

    scan_jobs = scan_jobs.by_order('dispatch_date')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'ascending'
    assert 'dispatch_date' in str(scan_jobs.order_by)
    assert [s.dispatch_date.day for s in scan_jobs.query()] == [1, 2, 3, 4]

    scan_jobs = scan_jobs.by_order('delivery_number')
    assert scan_jobs.current_sort_by == 'delivery_number'
    assert scan_jobs.current_sort_order == 'ascending'
    assert 'delivery_number' in str(scan_jobs.order_by)
    assert [s.delivery_number for s in scan_jobs.query()] == [1, 1, 1, 2]

    scan_jobs = scan_jobs.by_order('delivery_number')
    assert scan_jobs.current_sort_by == 'delivery_number'
    assert scan_jobs.current_sort_order == 'descending'
    assert 'delivery_number' in str(scan_jobs.order_by)
    assert 'DESC' in str(scan_jobs.order_by)
    assert [s.delivery_number for s in scan_jobs.query()] == [2, 1, 1, 1]

    scan_jobs = scan_jobs.by_order('municipality_id')
    assert scan_jobs.current_sort_by == 'municipality_id'
    assert scan_jobs.current_sort_order == 'ascending'
    assert 'name' in str(scan_jobs.order_by)
    assert [s.municipality.name for s in scan_jobs.query()] == [
        'First', 'Phirrrrd', 'Phirrrrd', 'Śecond'
    ]

    scan_jobs = scan_jobs.by_order('municipality_id')
    assert scan_jobs.current_sort_by == 'municipality_id'
    assert scan_jobs.current_sort_order == 'descending'
    assert 'name' in str(scan_jobs.order_by)
    assert 'DESC' in str(scan_jobs.order_by)
    assert [s.municipality.name for s in scan_jobs.query()] == [
        'Śecond', 'Phirrrrd', 'Phirrrrd', 'First'
    ]

    scan_jobs = scan_jobs.by_order(None)
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'

    scan_jobs = scan_jobs.by_order('')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'

    scan_jobs = scan_jobs.by_order('xxx')
    assert scan_jobs.current_sort_by == 'dispatch_date'
    assert scan_jobs.current_sort_order == 'descending'



from onegov.core.converters import extended_date_converter
from onegov.core.converters import uuid_converter
from onegov.user import Auth
from onegov.user import User
from onegov.user import UserCollection
from onegov.wtfs.app import WtfsApp
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.collections import NotificationCollection
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.models import DailyList
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import DailyListBoxesAndForms
from onegov.wtfs.models import Invoice
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import Notification
from onegov.wtfs.models import Principal
from onegov.wtfs.models import Report
from onegov.wtfs.models import ReportBoxes
from onegov.wtfs.models import ReportBoxesAndForms
from onegov.wtfs.models import ReportBoxesAndFormsByDelivery
from onegov.wtfs.models import ReportFormsByMunicipality
from onegov.wtfs.models import ScanJob
from onegov.wtfs.models import UserManual
from webob.exc import HTTPNotFound


@WtfsApp.path(
    model=Principal,
    path='/'
)
def get_principal(app):
    return app.principal


@WtfsApp.path(
    model=Auth,
    path='/auth'
)
def get_auth(request, to='/'):
    return Auth.from_request(request, to)


@WtfsApp.path(
    model=UserCollection,
    path='/users'
)
def get_users(request):
    return UserCollection(request.session)


@WtfsApp.path(
    model=User,
    path='/user/{username}'
)
def get_user(request, username):
    return UserCollection(request.session).by_username(username)


@WtfsApp.path(
    model=MunicipalityCollection,
    path='/municipalities'
)
def get_municipalities(request):
    return MunicipalityCollection(request.session)


@WtfsApp.path(
    model=Municipality,
    path='/municipality/{id}',
    converters=dict(
        id=uuid_converter
    )
)
def get_municipality(request, id):
    return MunicipalityCollection(request.session).by_id(id)


@WtfsApp.path(
    model=ScanJobCollection,
    path='/scan-jobs',
    converters=dict(
        page=int,
        from_date=extended_date_converter,
        to_date=extended_date_converter,
        type=[str],
        municipality_id=[str],
        sort_by=str,
        sort_order=str
    )
)
def get_scan_jobs(
    request,
    page=None,
    from_date=None,
    to_date=None,
    type=None,
    municipality_id=None,
    term=None,
    sort_by=None,
    sort_order=None
):
    return ScanJobCollection(
        request.session,
        page=page,
        group_id=(
            None if request.has_role('admin')
            else getattr(request.identity, 'groupid', 'nogroup')
        ),
        from_date=from_date,
        to_date=to_date,
        type=type,
        municipality_id=municipality_id,
        term=term,
        sort_by=sort_by,
        sort_order=sort_order
    )


@WtfsApp.path(
    model=ScanJob,
    path='/scan-job/{id}',
    converters=dict(
        id=uuid_converter
    )
)
def get_scan_job(request, id):
    return ScanJobCollection(request.session).by_id(id)


@WtfsApp.path(
    model=DailyList,
    path='/daily-list',
)
def get_daily_list(request):
    return DailyList()


@WtfsApp.path(
    model=DailyListBoxes,
    path='/daily-list/boxes/{date}',
    converters=dict(
        date=extended_date_converter,
    )
)
def get_daily_list_boxes(request, date):
    return DailyListBoxes(request.session, date)


@WtfsApp.path(
    model=DailyListBoxesAndForms,
    path='/daily-list/boxes-and-forms/{date}',
    converters=dict(
        date=extended_date_converter,
    )
)
def get_daily_list_boxes_and_forms(request, date):
    return DailyListBoxesAndForms(request.session, date)


@WtfsApp.path(
    model=Report,
    path='/report',
)
def get_report(request):
    return Report(request.session)


@WtfsApp.path(
    model=ReportBoxes,
    path='/report/boxes/{start}/{end}',
    converters=dict(
        start=extended_date_converter,
        end=extended_date_converter,
    )
)
def get_report_boxes(request, start, end):
    return ReportBoxes(request.session, start, end)


@WtfsApp.path(
    model=ReportBoxesAndForms,
    path='/report/boxes-and-forms/{start}/{end}/{type}',
    converters=dict(
        start=extended_date_converter,
        end=extended_date_converter,
    )
)
def get_report_boxes_and_forms(request, start, end, type):
    return ReportBoxesAndForms(request.session, start, end, type)


@WtfsApp.path(
    model=ReportFormsByMunicipality,
    path='/report/forms/{start}/{end}/{type}/{municipality_id}',
    converters=dict(
        start=extended_date_converter,
        end=extended_date_converter,
        municipality_id=uuid_converter
    )
)
def get_report_forms(request, start, end, type, municipality_id):
    if not municipality_id:
        raise HTTPNotFound()
    return ReportFormsByMunicipality(
        request.session, start, end, type, municipality_id
    )


@WtfsApp.path(
    model=ReportBoxesAndFormsByDelivery,
    path='/report/delivery/{start}/{end}/{type}/{municipality_id}',
    converters=dict(
        start=extended_date_converter,
        end=extended_date_converter,
        municipality_id=uuid_converter
    )
)
def get_report_delivery(request, start, end, type, municipality_id):
    if not municipality_id:
        raise HTTPNotFound()
    return ReportBoxesAndFormsByDelivery(
        request.session, start, end, type, municipality_id
    )


@WtfsApp.path(
    model=NotificationCollection,
    path='/notifications'
)
def get_notifications(request):
    return NotificationCollection(request.session)


@WtfsApp.path(
    model=Notification,
    path='/notification/{id}'
)
def get_notification(request, id):
    return NotificationCollection(request.session).by_id(id)


@WtfsApp.path(
    model=Invoice,
    path='/invoice',
)
def get_invoice(request):
    return Invoice(request.session)


@WtfsApp.path(
    model=PaymentTypeCollection,
    path='/payment-types'
)
def get_payment_types(request):
    return PaymentTypeCollection(request.session)


@WtfsApp.path(
    model=UserManual,
    path='/user-manual',
)
def get_user_manual(request):
    return UserManual(request.app)



from onegov.core.collection import GenericCollection
from onegov.core.orm.func import unaccent
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import PickupDate


class MunicipalityCollection(GenericCollection):

    @property
    def model_class(self):
        return Municipality

    def query(self):
        query = super(MunicipalityCollection, self).query()
        query = query.order_by(None).order_by(unaccent(Municipality.name))
        return query

    def import_data(self, data):
        for bfs_number, values in data.items():
            query = self.query()
            query = query.filter(Municipality.meta['bfs_number'] == bfs_number)
            municipality = query.first()
            if not municipality:
                continue

            dates = set(values['dates'])
            dates -= {d.date for d in municipality.pickup_dates}
            for date in dates:
                municipality.pickup_dates.append(PickupDate(date=date))



from morepath.request import Response
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.forms import CreateInvoicesForm
from onegov.wtfs.layouts import InvoiceLayout
from onegov.wtfs.models import Invoice
from onegov.wtfs.security import ViewModel


@WtfsApp.form(
    model=Invoice,
    template='form.pt',
    permission=ViewModel,
    form=CreateInvoicesForm
)
def create_invoices(self, request, form):
    """ Create invoices and download them as CSV. """

    layout = InvoiceLayout(self, request)

    if form.submitted(request):
        form.update_model(self)

        response = Response(
            content_type='text/csv',
            content_disposition='inline; filename=rechnungen.csv'
        )
        self.export(response.body_file)
        return response

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Create invoice"),
        'cancel': layout.cancel_url
    }



from chameleon import PageTemplate
from wtforms.fields import Field
from wtforms.widgets.core import html_params
from wtforms.widgets.core import HTMLString


class HintWidget(object):

    def __call__(self, field, **kwargs):
        params = html_params(name=field.name, **kwargs)
        request = field.meta.request
        registry = request.app.config.template_engine_registry
        macros = registry._template_loaders['.pt'].macros
        template = PageTemplate(
            f"""
                <div {params}>
                    <div metal:use-macro="macros['{field.macro}']"/>
                </div>
            """,
            macros=macros
        )

        translate = request.get_translate(for_chameleon=True)
        return HTMLString(template.render(translate=translate))


class HintField(Field):

    widget = HintWidget()

    def __init__(self, *args, **kwargs):
        self.macro = kwargs.pop('macro')
        super().__init__(*args, **kwargs)

    def populate_obj(self, obj, name):
        pass



from onegov.chat import MessageCollection


class NotificationCollection(MessageCollection):

    def __init__(self, session, **kwargs):
        kwargs.pop('type', None)
        super().__init__(session, type='wtfs_notification', **kwargs)



from cached_property import cached_property
from onegov.wtfs.models.report import Report


class DailyList(object):
    pass


class DailyListBoxes(Report):
    """ All boxes of all municipalities on one day. """

    def __init__(self, session, date_=None):
        super().__init__(session, date_, date_)
        self.date = self.start
        self.type = 'normal'

    @cached_property
    def columns_dispatch(self):
        return [
            'dispatch_boxes',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
        ]

    @cached_property
    def columns_return(self):
        return ['return_boxes']


class DailyListBoxesAndForms(Report):
    """ All boxes of all municipalities on one day. """

    def __init__(self, session, date_=None):
        super().__init__(session, date_, date_)
        self.date = self.start
        self.type = 'normal'

    @cached_property
    def columns_dispatch(self):
        return [
            'dispatch_boxes',
            'dispatch_tax_forms_older',
            'dispatch_tax_forms_last_year',
            'dispatch_tax_forms_current_year',
            'dispatch_single_documents',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
        ]



from onegov.form import Form
from onegov.form.fields import UploadField
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.wtfs import _


class UserManualForm(Form):

    pdf = UploadField(
        label=_("PDF"),
        validators=[
            WhitelistedMimeType({'application/pdf'}),
            FileSizeLimit(50 * 1024 * 1024)
        ]
    )

    def update_model(self, model):
        action = getattr(self.pdf, 'action', '')
        if action == 'delete':
            del model.pdf
        if action == 'replace':
            if self.pdf.data:
                model.pdf = self.pdf.raw_data[-1].file.read()

    def apply_model(self, model):
        if model.exists:
            self.pdf.data = {
                'filename': model.filename,
                'size': model.content_length,
                'mimetype': model.content_type
            }



class UserManual(object):

    def __init__(self, app):
        self.app = app
        self.filename = 'user_manual.pdf'
        self.content_type = 'application/pdf'

    @property
    def exists(self):
        return self.app.filestorage.exists(self.filename)

    @property
    def pdf(self):
        if self.exists:
            with self.app.filestorage.open(self.filename, 'rb') as file:
                result = file.read()
            return result

    @pdf.setter
    def pdf(self, value):
        with self.app.filestorage.open(self.filename, 'wb') as file:
            file.write(value)

    @pdf.deleter
    def pdf(self):
        if self.exists:
            self.app.filestorage.remove(self.filename)

    @property
    def content_length(self):
        if self.exists:
            return self.app.filestorage.getsize(self.filename)



from freezegun import freeze_time
from onegov.core.request import CoreRequest
from unittest.mock import patch
from webtest.forms import Upload


def test_views_invoices(client):
    # Add a municipality with dates
    client.login_admin()

    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '1'
    add.form['gpn_number'] = '11223344'
    add.form['payment_type'] = 'normal'
    assert "Adlikon" in add.form.submit().follow()

    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde;1;-1;Normal;5.1.2019".encode('cp1252'),
        'text/csv'
    )
    assert "Gemeindedaten importiert." in upload.form.submit().follow()

    # Add a scan job
    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs/unrestricted').click(href='/add')
        add.form['type'].select("normal")
        add.form['municipality_id'].select(text="Adlikon (1)")
        add.form['dispatch_date'] = "2019-01-05"
        add.form['dispatch_tax_forms_older'] = "30"
        add.form['dispatch_tax_forms_last_year'] = "20"
        add.form['dispatch_tax_forms_current_year'] = "10"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().maybe_follow()

        edit = client.get('/scan-jobs/unrestricted')\
            .click("05.01.2019").click("Bearbeiten")
        edit.form['return_date'] = "2019-01-10"
        edit.form['return_unscanned_tax_forms_older'] = "3"
        edit.form['return_unscanned_tax_forms_last_year'] = "2"
        edit.form['return_unscanned_tax_forms_current_year'] = "1"

        assert "Scan-Auftrag geändert." in edit.form.submit().maybe_follow()

    with freeze_time("2019-01-02"):
        create = client.get('/invoice')
        create.form['from_date'] = "2019-01-01"
        create.form['to_date'] = "2019-01-07"
        create.form['cs2_user'] = "123"
        create.form['subject'] = "Rechnungen"
        create.form['municipality_id'].select(text="Adlikon (1)")
        create.form['accounting_unit'] = "456"
        create.form['revenue_account'] = "789"
        created = create.form.submit()
        assert created.headers['Content-Type'] == 'text/csv; charset=UTF-8'
        assert created.headers['Content-Disposition'] == (
            'inline; filename=rechnungen.csv'
        )

        invoice = created.text
        assert "11223344" in invoice
        assert "Rechnungen" in invoice
        assert "123" in invoice
        assert "456" in invoice
        assert "789" in invoice
        assert "2019-01-02" in invoice
        assert "-00000000000006000" in invoice
        assert "+00000000070000000" in invoice


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_invoice_permissions(mock_method, client):
    client.get('/invoice', status=403)

    client.login_member()
    client.get('/invoice', status=403)
    client.logout()

    client.login_editor()
    client.get('/invoice', status=403)
    client.logout()

    client.login_admin()
    client.get('/invoice')
    client.logout()



from morepath import redirect
from onegov.core.templates import render_template
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.forms import AddScanJobForm
from onegov.wtfs.forms import EditScanJobForm
from onegov.wtfs.forms import ScanJobsForm
from onegov.wtfs.forms import UnrestrictedScanJobForm
from onegov.wtfs.forms import UnrestrictedScanJobsForm
from onegov.wtfs.layouts import AddScanJobLayout
from onegov.wtfs.layouts import DeliveryNoteLayout
from onegov.wtfs.layouts import EditScanJobLayout
from onegov.wtfs.layouts import MailLayout
from onegov.wtfs.layouts import ScanJobLayout
from onegov.wtfs.layouts import ScanJobsLayout
from onegov.wtfs.models import ScanJob
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import AddModelUnrestricted
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import EditModelUnrestricted
from onegov.wtfs.security import ViewModel
from onegov.wtfs.security import ViewModelUnrestricted


@WtfsApp.form(
    model=ScanJobCollection,
    permission=ViewModel,
    form=ScanJobsForm,
    template='scan_jobs.pt'
)
def view_scan_jobs(self, request, form):
    if request.has_permission(self, ViewModelUnrestricted):
        return redirect(request.link(self, 'unrestricted'))

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': ScanJobsLayout(self, request),
        'form': form,
        'permission': ViewModel
    }


@WtfsApp.form(
    model=ScanJobCollection,
    name='unrestricted',
    permission=ViewModelUnrestricted,
    form=UnrestrictedScanJobsForm,
    template='scan_jobs_unrestricted.pt'
)
def view_scan_jobs_unrestricted(self, request, form):
    if not form.errors:
        form.apply_model(self)

    return {
        'layout': ScanJobsLayout(self, request),
        'form': form,
        'permission': ViewModel
    }


@WtfsApp.form(
    model=ScanJobCollection,
    name='add-unrestricted',
    template='form.pt',
    permission=AddModelUnrestricted,
    form=UnrestrictedScanJobForm
)
def add_scan_job_unrestricted(self, request, form):
    """ Create a new scan job. """
    layout = AddScanJobLayout(self, request)

    if form.submitted(request):
        scan_job = ScanJob()
        form.update_model(scan_job)
        scan_job.delivery_number = self.next_delivery_number(
            scan_job.municipality_id
        )
        request.session.add(scan_job)
        request.message(_("Scan job added."), 'success')
        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }


@WtfsApp.form(
    model=ScanJobCollection,
    name='add',
    template='form.pt',
    permission=AddModel,
    form=AddScanJobForm
)
def add_scan_job(self, request, form):
    """ Create a new scan job. """

    if request.has_permission(self, AddModelUnrestricted):
        return redirect(request.link(self, name='add-unrestricted'))

    layout = AddScanJobLayout(self, request)

    if form.submitted(request):
        scan_job = ScanJob()
        form.update_model(scan_job)
        scan_job.delivery_number = self.next_delivery_number(
            scan_job.municipality_id
        )
        request.session.add(scan_job)
        request.message(_("Scan job added."), 'success')

        subject = request.translate(
            _("Order confirmation for scanning your tax returns")
        )
        request.app.send_transactional_email(
            subject=subject,
            receivers=(request.identity.userid, ),
            reply_to=request.app.mail['transactional']['sender'],
            content=render_template(
                'mail_confirm.pt',
                request,
                {
                    'title': subject,
                    'model': scan_job,
                    'layout': MailLayout(scan_job, request)
                }
            ),
        )

        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }


@WtfsApp.html(
    model=ScanJob,
    template='scan_job.pt',
    permission=ViewModel
)
def view_scan_job(self, request):
    """ View a single scan job. """
    layout = ScanJobLayout(self, request)

    return {
        'layout': layout,
    }


@WtfsApp.form(
    model=ScanJob,
    name='edit-unrestricted',
    template='form.pt',
    permission=EditModelUnrestricted,
    form=UnrestrictedScanJobForm
)
def edit_scan_job_unrestricted(self, request, form):
    """ Edit a scan job. """

    layout = EditScanJobLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Scan job modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.form(
    model=ScanJob,
    name='edit',
    template='form.pt',
    permission=EditModel,
    form=EditScanJobForm
)
def edit_scan_job(self, request, form):
    """ Edit a scan job. """

    if request.has_permission(self, EditModelUnrestricted):
        return redirect(request.link(self, name='edit-unrestricted'))

    layout = EditScanJobLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Scan job modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'subtitle': layout.format_date(self.dispatch_date, 'date'),
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.view(
    model=ScanJob,
    request_method='DELETE',
    permission=DeleteModel
)
def delete_scan_job(self, request):
    """ Delete a scan job. """

    request.assert_valid_csrf_token()
    ScanJobCollection(request.session).delete(self)
    request.message(_("Scan job deleted."), 'success')


@WtfsApp.html(
    model=ScanJob,
    template='delivery_note.pt',
    name='delivery-note',
    permission=ViewModel
)
def view_scan_job_delivery_note(self, request):
    """ View a single scan job. """
    layout = DeliveryNoteLayout(self, request)

    return {
        'layout': layout,
    }



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.wtfs import _
from onegov.wtfs.layouts.default import DefaultLayout
from onegov.wtfs.security import EditModel


class UserManualLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("User manual")

    @cached_property
    def editbar_links(self):
        result = []
        if self.request.has_permission(self.model, EditModel):
            result.append(
                Link(
                    text=_("Edit"),
                    url=self.request.link(self.model, 'edit'),
                    attrs={'class': 'edit-icon'}
                )
            )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, '#')
        ]


class EditUserManualLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Edit user manual")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("User manual"), self.user_manual_url),
            Link(_("Edit"), '#')
        ]

    @cached_property
    def cancel_url(self):
        return self.user_manual_url

    @cached_property
    def success_url(self):
        return self.user_manual_url



from cached_property import cached_property
from onegov.core import Framework
from onegov.core import utils
from onegov.file import DepotApp
from onegov.form import FormApp
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.models import Principal
from onegov.wtfs.models import PaymentType
from onegov.wtfs.theme import WtfsTheme


class WtfsApp(Framework, FormApp, DepotApp):
    """ The Wtfs application. Include this in your onegov.yml to serve
    it with onegov-server.

    """

    serve_static_files = True

    @cached_property
    def principal(self):
        return Principal()

    def add_initial_content(self):
        session = self.session()
        municipalities = MunicipalityCollection(session)
        for name, bfs_number in (
            ('Adlikon', 21),
            ('Aesch', 241),
            ('Aeugst am Albis', 1),
            ('Altikon', 211),
            ('Andelfingen', 30),
            ('Bachenbülach', 51),
            ('Bachs', 81),
            ('Bäretswil', 111),
            ('Bauma', 171),
            ('Benken', 22),
            ('Berg am Irchel', 23),
            ('Bertschikon', 212),
            ('Birmensdorf', 242),
            ('Bonstetten', 3),
            ('Boppelsen', 82),
            ('Brütten', 213),
            ('Buch am Irchel', 24),
            ('Dachsen', 25),
            ('Dägerlen', 214)
        ):
            municipalities.add(
                name=name,
                bfs_number=bfs_number
            )
        session.add(PaymentType(name='normal', _price_per_quantity=700))
        session.add(PaymentType(name='spezial', _price_per_quantity=850))


@WtfsApp.static_directory()
def get_static_directory():
    return 'static'


@WtfsApp.template_directory()
def get_template_directory():
    return 'templates'


@WtfsApp.setting(section='core', name='theme')
def get_theme():
    return WtfsTheme()


@WtfsApp.setting(section='i18n', name='localedirs')
def get_i18n_localedirs():
    return [
        utils.module_path('onegov.wtfs', 'locale'),
        utils.module_path('onegov.form', 'locale'),
        utils.module_path('onegov.user', 'locale')
    ]


@WtfsApp.setting(section='i18n', name='locales')
def get_i18n_used_locales():
    return {'de_CH'}


@WtfsApp.setting(section='i18n', name='default_locale')
def get_i18n_default_locale():
    return 'de_CH'


@WtfsApp.webasset_path()
def get_shared_assets_path():
    return utils.module_path('onegov.shared', 'assets/js')


@WtfsApp.webasset_path()
def get_js_path():
    return 'assets/js'


@WtfsApp.webasset_path()
def get_css_path():
    return 'assets/css'


@WtfsApp.webasset_output()
def get_webasset_output():
    return 'assets/bundles'


@WtfsApp.webasset('frameworks')
def get_frameworks_asset():
    yield 'modernizr.js'
    yield 'jquery.js'
    yield 'jquery.tablesorter.js'
    yield 'tablesaw.css'
    yield 'tablesaw.jquery.js'
    yield 'tablesaw-create.js'
    yield 'tablesaw-init.js'
    yield 'foundation.js'
    yield 'intercooler.js'
    yield 'underscore.js'
    yield 'react.js'
    yield 'react-dom.js'
    yield 'form_dependencies.js'
    yield 'confirm.jsx'
    yield 'jquery.datetimepicker.css'
    yield 'jquery.datetimepicker.js'
    yield 'datetimepicker.js'


@WtfsApp.webasset('common')
def get_common_asset():
    yield 'common.js'



from onegov.core.security import Public
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.layouts import DefaultLayout
from webob.exc import HTTPForbidden
from webob.exc import HTTPNotFound


@WtfsApp.html(
    model=HTTPForbidden,
    template='exception.pt',
    permission=Public
)
def handle_forbidden(self, request):
    """ Displays a nice HTTP 403 error. """

    @request.after
    def set_status_code(response):
        response.status_code = self.code

    return {
        'layout': DefaultLayout(self, request),
        'title': _("Access Denied"),
        'message': _(
            "You are trying to open a page for which you are not authorized."
        )
    }


@WtfsApp.html(
    model=HTTPNotFound,
    template='exception.pt',
    permission=Public
)
def handle_notfound(self, request):
    """ Displays a nice HTTP 404 error. """

    @request.after
    def set_status_code(response):
        response.status_code = self.code

    return {
        'layout': DefaultLayout(self, request),
        'title': _("Page not Found"),
        'message': _("The page you are looking for could not be found."),
    }



from cgi import FieldStorage
from datetime import date
from io import BytesIO
from onegov.core.utils import Bunch
from onegov.form import Form
from onegov.wtfs.fields import HintField
from onegov.wtfs.fields import MunicipalityDataUploadField


class PostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


def test_municipality_data_upload_field():
    form = Form()

    def process(content, **kwargs):
        field = MunicipalityDataUploadField(**kwargs)
        field = field.bind(form, 'upload')

        field_storage = FieldStorage()
        field_storage.file = BytesIO(content)
        field_storage.type = 'text/plain'
        field_storage.filename = 'test.csv'

        field.process(PostData({'upload': field_storage}))
        return field

    # Invalid
    field = process('Bäretswil;\r\n'.encode('cp1252'))
    assert not field.validate(form)
    errors = [error.interpolate() for error in field.errors]
    assert "Some rows contain invalid values: 0." in errors

    # Valid
    field = process('Bäretswil;111;-1;Normal;\r\n'.encode('cp1252'))
    assert field.validate(form)
    assert field.data == {111: {'dates': []}}

    field = process(
        'Bäretswil;111;-1;Normal;01.01.2019;07.01.2019\r\n'.encode('cp1252')
    )
    assert field.validate(form)
    assert field.data == {
        111: {'dates': [date(2019, 1, 1), date(2019, 1, 7)]}
    }


def test_hint_field(wtfs_app):
    def get_translate(for_chameleon):
        return wtfs_app.chameleon_translations.get('de_CH')

    form = Form()
    field = HintField(macro='express_shipment_hint')
    field = field.bind(form, 'hint')
    field.meta.request = Bunch(app=wtfs_app, get_translate=get_translate)

    assert field.validate(form)
    assert "Für dringende Scan-Aufträge" in field()



from humanize import naturalsize
from morepath import redirect
from morepath import Response
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.forms import UserManualForm
from onegov.wtfs.layouts import EditUserManualLayout
from onegov.wtfs.layouts import UserManualLayout
from onegov.wtfs.models import UserManual
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import ViewModel


@WtfsApp.html(
    model=UserManual,
    template='user_manual.pt',
    permission=ViewModel
)
def view_user_manual(self, request):
    """ View the user manual. """
    layout = UserManualLayout(self, request)
    filesize = naturalsize(self.content_length or 0)

    return {
        'layout': layout,
        'filesize': filesize
    }


@WtfsApp.form(
    model=UserManual,
    name='edit',
    template='form.pt',
    permission=EditModel,
    form=UserManualForm
)
def edit_user_manual(self, request, form):
    """ Edit the user manual """

    layout = EditUserManualLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("User manual modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.view(
    model=UserManual,
    permission=ViewModel,
    name='pdf'
)
def user_manual_pdf(self, request):
    if not self.pdf:
        return Response(status='503 Service Unavailable')

    return Response(
        self.pdf,
        content_type='application/pdf',
        content_disposition='inline; filename={}.pdf'.format(
            request.translate(_("User manual"))
        )
    )



from csv import writer
from datetime import datetime
from onegov.wtfs.models.municipality import Municipality
from onegov.wtfs.models.scan_job import ScanJob


class Invoice(object):

    def __init__(self, session):
        self.session = session
        self.from_date = None
        self.to_date = None
        self.cs2_user = None
        self.subject = None
        self.municipality_id = None
        self.accounting_unit = None
        self.revenue_account = None

    def export(self, file):
        csv = writer(file, delimiter=';')
        now = datetime.now()
        current_date = f'{now:%Y-%m-%d}'
        current_time = f'{now:%H.%M.%S}'

        query = self.session.query(ScanJob)
        query = query.join(Municipality)
        if self.from_date:
            query = query.filter(ScanJob.dispatch_date >= self.from_date)
        if self.to_date:
            query = query.filter(ScanJob.dispatch_date <= self.to_date)
        if self.municipality_id:
            query = query.filter(
                ScanJob.municipality_id == self.municipality_id
            )
        query = query.order_by(
            Municipality.meta['gpn_number'],
            ScanJob.delivery_number
        )

        item_number = 1
        municipality_count = 0
        last_municipality = None
        for scan_job in query:
            municipality = scan_job.municipality
            if municipality == last_municipality:
                item_number += 1
            else:
                item_number = 1
                municipality_count += 1

            csv.writerow((
                f'{now:%Y%m%d}{municipality_count}',
                current_date,
                current_time,
                f'{municipality.gpn_number:08}',
                f'{item_number:05}',
                f'{municipality.gpn_number:08}',
                self.subject,
                f'{item_number:05}',
                f'{item_number:03}',
                current_date,
                current_date,
                current_date,
                f'{municipality.gpn_number:08}',
                f'{int((scan_job.return_scanned_tax_forms or 0) * 1000):+018}',
                f'Lieferscheinnummer {scan_job.delivery_number}',
                f'{int(municipality.price_per_quantity * 10000000):+018}',
                f'{int(municipality.price_per_quantity * 10000000):+018}',
                '1',
                current_date,
                municipality.address_supplement or '',
                self.cs2_user,
                '000',
                self.accounting_unit,
                self.revenue_account,
            ))

            last_municipality = municipality



from cached_property import cached_property
from datetime import date
from onegov.core.elements import Link
from onegov.core.layout import ChameleonLayout
from onegov.user import Auth
from onegov.user import UserCollection
from onegov.wtfs import _
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.collections import NotificationCollection
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.models import DailyList
from onegov.wtfs.models import Invoice
from onegov.wtfs.models import Report
from onegov.wtfs.models import UserManual
from onegov.wtfs.security import ViewModel


class DefaultLayout(ChameleonLayout):

    day_long_format = 'skeleton:MMMMd'
    date_long_format = 'long'
    datetime_long_format = 'medium'

    def __init__(self, model, request):
        super().__init__(model, request)
        self.request.include('frameworks')
        self.request.include('chosen')
        self.request.include('common')

    @cached_property
    def title(self):
        return ""

    @cached_property
    def top_navigation(self):
        has_permission = self.request.has_permission
        session = self.request.session
        result = []
        if has_permission(ScanJobCollection(session), ViewModel):
            result.append(Link(_("Scan jobs"), self.scan_jobs_url))
        if has_permission(DailyList(), ViewModel):
            result.append(Link(_("Daily list"), self.daily_list_url))
        if has_permission(Report(session), ViewModel):
            result.append(Link(_("Report"), self.report_url))
        if has_permission(Invoice(session), ViewModel):
            result.append(Link(_("Invoices"), self.invoices_url))
        if has_permission(UserCollection(session), ViewModel):
            result.append(Link(_("Users"), self.users_url))
        if has_permission(MunicipalityCollection(session), ViewModel):
            result.append(Link(_("Municipalities"), self.municipalities_url))
        if has_permission(NotificationCollection(session), ViewModel):
            result.append(Link(_("Notifications"), self.notifications_url))
        if has_permission(UserManual(self.app), ViewModel):
            result.append(Link(_("User manual"), self.user_manual_url))

        return result

    @cached_property
    def editbar_links(self):
        return []

    @cached_property
    def breadcrumbs(self):
        return [Link(_("Homepage"), self.homepage_url)]

    @cached_property
    def app_version(self):
        return self.app.settings.core.theme.version

    @cached_property
    def static_path(self):
        return self.request.link(self.app.principal, 'static')

    @cached_property
    def homepage_url(self):
        return self.request.link(self.app.principal)

    @cached_property
    def login_url(self):
        if not self.request.is_logged_in:
            return self.request.link(
                Auth.from_request(self.request, to=self.homepage_url),
                name='login'
            )

    @cached_property
    def logout_url(self):
        if self.request.is_logged_in:
            return self.request.link(
                Auth.from_request(self.request, to=self.homepage_url),
                name='logout'
            )

    @cached_property
    def users_url(self):
        return self.request.link(UserCollection(self.request.session))

    @cached_property
    def municipalities_url(self):
        return self.request.link(MunicipalityCollection(self.request.session))

    @cached_property
    def scan_jobs_url(self):
        return self.request.link(ScanJobCollection(self.request.session))

    @cached_property
    def daily_list_url(self):
        return self.request.link(DailyList())

    @cached_property
    def report_url(self):
        return self.request.link(Report(self.request.session))

    @cached_property
    def invoices_url(self):
        return self.request.link(Invoice(self.request.session))

    @cached_property
    def notifications_url(self):
        return self.request.link(NotificationCollection(self.request.session))

    @cached_property
    def payment_types_url(self):
        return self.request.link(PaymentTypeCollection(self.request.session))

    @cached_property
    def user_manual_url(self):
        return self.request.link(UserManual(self.app))

    @cached_property
    def cancel_url(self):
        return ''

    @cached_property
    def success_url(self):
        return ''

    @cached_property
    def current_year(self):
        return date.today().year

    @property
    def notifications(self):
        return NotificationCollection(self.request.session)
